/* Programs takes 4GL file in ingres format and 
   1) saves it in format
      suitable for editing with expand macro code and debugging code
      removed.  
OR    
   2) saves it in with macros expanded and debugging code in place
      suitable for ingres compilation
OR
   3) scans file only creating symbols for each include file
      and # of includes

   Format osqedit mode input_file output_file.

   V1.2c  10/7/91
   V1.3c  3/3/91
	Modified to recognise ';' as a delimiter for called procedure
        parameters.  Previously ';' taken to mean statement delimiter.
        New assumption ';' nested inside parentheses is no longer treated
        as statement delimiter.
        Added better handling of WHILE..DO constructs.
	Added better handling of ELSE construct.
   V1.4c  3/3/91
        Modified debug handler to ensure that top line of debug statement
        is not blank.  This was because in POPUP=OFF mode only 1 line of
        debug statement is displayed and this was sometimes blank.
        Added sleep command macro that allows the user to alter or disable
        the sleep command after POPUP=OFF message.  Valid values are 0/OFF,
        1,2,3,4
 
  V2.0c   5/12/92
        Support macro parameter in {} as well as <>.
        Fixed ; delimiter where , delimiter should have been used (Vax 
        compiler did not mind! ). All messages written to stderr not stdout.

  V2.0ux  5/12/92
     UNIX VERSION DG/UX 5.4 VERSION
        Modified to distinguish upper and lowercase filenames.
        Use standard unix fileio. Environment variables used rather than
        symbols for passing back file list. Environment variables not set in
        C code but list written out to stdout and grabbed by script for
        processing.  Created script file to replace osqedit.com.
  V2.1ux 6/30/92
        Fixed bug i<=10 should have been i<=9. Was overrunning array and
        causing corruption or failure on some machines.

 
   Commands recognised
     Note parameters may be nested in <> or {} to support
     debug conditions that use less than or greater than operators.

           / * #include <filename> * /
                insert filename specified into source file immediately
                after this include statement end comment mark.
           / * #debug <on>|<off>|<4gl condition> * / (default=off)
           / * #watch <4gl variable name> * /
           / * #nowatch <4gl variable name> * /
           / * #popup <on>|<off> * /                 (default=on)
           / * #sleep <0>|<off>|<1>|<2>|<3>|<4> * /  (default=1)

   1) Note that the macro command must be enclosed in comments so that Ingres
      does not treat this as one of its own commands.
   2) A single space must be between start comment * and # sign.  Ingres treats
      / * # as a special situation and will generate weird errors if you use
      this form.

   3) Limits no statement can be > 30000 characters
             no word can be > 1000 characters
             no macro command can be > 1000 characters
	     no macro parameter can be > 1000 characters
	     no more than 50 different include files
	     include filenames cannot be longer than 100 characters
             condition debug statement cannot be longer than 1000 characters
	     program cannot nest blocks more than 50 levels deep
             program cannot nest include files more than 10 levels deep
	     no more than 50 variables can be watched simultaneously
*/


#include	<stdio.h>

#define	TRUE		1
#define	FALSE		0
#define BUFSIZE		512


FILE		*infile,*outfile;
char		current_statement[30000];
char            actual_statement[30000];
char		print_statement[30000];
char            macro_command[1000];
char            macro_param[200],macro_param_upper[200];

char		current_char,sleep_string[20];
char		current_word[1000];
char            filelist[50][100];
char        	condit_dbg[1000];
char 		watchlist[30][200];
int             quote_flag,hash_flag;
int         	comment_flag,debug_flag,err_flag;
int  		inccount,watchcnt,watchok;
int             paren_level;                       /* V1.3c */

int         	result,popmode,fileok,status;
int         	i,j,ignore;
int         	level,countlev[50];
FILE         	*incfile[50];
char        	file_err[]="/* #!!! INCLUDE file not found !!!*/";
char        	sym_name [80];
char        	sym_value [100] ;
char        	*statpnt,mode_flag;

char 		writebuf[BUFSIZ+1];
char            *writepnt;
char            readbuf[10][BUFSIZ+1];
char            *readpnt[10];


extern void     strupr();
extern char     mjgetc();
extern int      next_statement();
extern void     process_i ();
extern void     process_e ();
extern void     process_s ();
extern int      mjflush ();
extern int      mjfputs ();

/* Get arguments */
/*************************************************************************/
 main(argc, argv)
/*************************************************************************/
int  argc;
char **argv;
{

    if (argc!=4)         /* Must be 3 arguments + program name */ 
    {
         perror ("Format: osqedit mode input_file output_file \n\n");
	 exit(1);
    };

    argv++;		/* skip program name */
                	/* 1st argument */
    if(*argv)           /* First argument 'e', 's' or 'i' */
    {
     mode_flag=tolower(**argv);
     if(mode_flag!='e'&&mode_flag!='s'&&mode_flag!='i')
     {
       perror(*argv);
       perror("Mode must be specified as 1st argument  e,s or i");
       exit(1);
     }
    };
    argv++;



                /* 2nd argument */
    if(*argv)
    {	
                /* open file in read mode */
         infile=fopen(*argv,"r");
	 if(infile == NULL )
	   {
	     perror(*argv);
	     exit(1);
	   };
    };
    argv++;

if(mode_flag!='i')
 {
                /* 3rd argument */
	if(*argv)
	{
	  outfile = fopen(*argv,"w");
	  if(outfile==NULL)
	   {
             perror(*argv);                  /* Error so print file name */
	     exit(1);
	   };
     };
 };

                                         /* Initialize everything */
    condit_dbg[0]='\0';
    watchcnt=0;
    debug_flag=0;
    popmode=1;
    current_char='\0';
    inccount=0;
    incfile[0]=infile;
    level=0;
    countlev[0]=0;
    ignore=0;
    result=255;
                                      /* Setup pointers */
for (i=0; i<=9; i++)                  /* V2.1 */
     {	
      readpnt[i]=(&readbuf[i][0]);
      *readpnt[i]='\0';
     };
writepnt=writebuf;

    if(mode_flag=='i') process_i();	
    if(mode_flag=='e') process_e();    /* Files ok so process them */
    if(mode_flag=='s') process_s();
    fclose(infile);
    if(mode_flag!='i')mjflush();
    if(mode_flag!='i')fclose(outfile);
    exit(0);
}



/* PROCESS FILE FOR SCANNING */
/*************************************************************************/
/* Process the file specified in infile and figure all unique include
   files and store in symbol.  Same as editing processing except no
   output file written. */
 void process_i()
/*************************************************************************/
{


     /* MAIN LOOP BEGIN */
 while(result!=0)                        /* Loop until not end of file */
 {
  result=next_statement(0);              /* Get 1 4gl statement */

  err_flag=0;                            /* Assume no errors */
  if(result==10)                         /* Was it macro command? */
  {
  strupr(macro_command,macro_command);
  strcpy(macro_param_upper,macro_param);
  strupr(macro_param_upper,macro_param_upper);
/*********************************************************************/
  if(strcmp(macro_command,"INCLUDE")==0) /* Yes was it include? */
   {
    fileok=0;                            /* Assume file not yet checked */
    if(inccount>0)                       /* Look through files checked */
    {                                    /*   if any exist */
     for (i=1; i<=inccount; i++)  
     {
      if(strcmp(filelist[i],macro_param)==0)fileok=1; /* File found */
     };
    };

    if(fileok==0)                        /* File not found so check it */
    {                                    /* Try to open file */
         incfile[inccount+1]=fopen(macro_param,"r");
	 if( incfile[inccount+1] == NULL ) 
	   {
         perror(macro_param);            /* Error so print file name */
         err_flag=1;                     /* Set error flag */
	   }
         else
           {
            inccount++;                     /* No error so save file handle */
            strcpy(&filelist[inccount][0],macro_param); /* Save file name */



            strcpy(sym_value,macro_param);           /* Make env var i# which */
                                                     /*  contains this filename */
            printf("i%-d=%s\n",inccount,sym_value);
            printf("export i%-d\n",inccount);
           };    
    };
   };


/*********************************************************************/
                /* Remove old file error message.  This will be recreated
                   if there is still an error opening the include file.
                   flag ignore=2 will suppress writing to the output file
                   for one statement only */

  if(strcmp(macro_command,"!!!INCLUDEFILENOTFOUND!!!/")==0)
   {
     if(ignore==0)ignore=2;  /* No need to suppress output if all output
                                already suppressed */
   };

/*********************************************************************/
                 /* Ignore all statements between DEBUGBEG and DEBUGEND */
  if(strcmp(macro_command,"DEBUGBEG/")==0)
     {
     ignore=1;
   };
  if(strcmp(macro_command,"DEBUGEND/")==0)
   {
     if(level==0)ignore=2;
   };
/*********************************************************************/
                 /* Ignore all statements between MACROBEG and MACROEND */
  if(strcmp(macro_command,"MACROBEG/")==0)
   {
     level++;
     ignore=1;
   };
  if(strcmp(macro_command,"MACROEND/")==0)
   {
     level--;
     if(level==0)ignore=2;
   };
  };
/*********************************************************************/

  if(ignore==2)ignore=0;

/*********************************************************************/

};    /* MAIN LOOP END */

                    /* Setup env var  include with # distinct
                       include files */
printf("includes=%-d\n",inccount);
printf("export includes\n");
};









/* PROCESS FILE FOR EDITING */
/*************************************************************************/
/* Process the file specified in infile remove all expanded include files,
   debug information and old error information and write file to outfile */
 void process_e()
/*************************************************************************/
{

     /* MAIN LOOP BEGIN */
 while(result!=0)                        /* Loop until not end of file */
 {
  result=next_statement(0);              /* Get 1 4gl statement */
  err_flag=0;                            /* Assume no errors */
  if(result==10)                         /* Was it macro command? */
  {
  strupr(macro_command,macro_command);
  strcpy(macro_param_upper,macro_param);
  strupr(macro_param_upper,macro_param_upper);
/*********************************************************************/
  if(strcmp(macro_command,"INCLUDE")==0) /* Yes was it include? */
   {
    fileok=0;                            /* Assume file not yet checked */
    if(inccount>0)                       /* Look through files checked */
    {                                    /*   if any exist */
     for (i=1; i<=inccount; i++)  
     {
      if(strcmp(macro_param,&filelist[i][0])==0)fileok=1; /* File found */
     };
    };

    if(fileok==0)                        /* File not found so check it */
    {                                    /* Try to open file */
         incfile[inccount+1]=fopen(macro_param,"r");
	 if( incfile[inccount+1] == NULL)
	   {
         perror(macro_param);            /* Error so print file name */
         err_flag=1;                     /* Set error flag */
	   }
         else
           {
            inccount++;                     /* No error so save file handle */
            strcpy(&filelist[inccount][0],macro_param); /* Save file name */

            strcpy(sym_value,macro_param);           /* Make env var i# which */
                                                     /*  contains this filename */
            printf("i%-d=%s\n",inccount,sym_value);
            printf("export i%-d\n",inccount);
           };    
    };
   };


/*********************************************************************/
                /* Remove old file error message.  This will be recreated
                   if there is still an error opening the include file.
                   flag ignore=2 will suppress writing to the output file
                   for one statement only */

  if(strcmp(macro_command,"!!!INCLUDEFILENOTFOUND!!!/")==0)
   {
     if(ignore==0)ignore=2;  /* No need to suppress output if all output
                                already suppressed */
   };

/*********************************************************************/
                 /* Ignore all statements between DEBUGBEG and DEBUGEND */
  if(strcmp(macro_command,"DEBUGBEG/")==0)
     {
     ignore=1;
   };
  if(strcmp(macro_command,"DEBUGEND/")==0)
   {
     if(level==0)ignore=2;
   };
/*********************************************************************/
                 /* Ignore all statements between MACROBEG and MACROEND */
  if(strcmp(macro_command,"MACROBEG/")==0)
   {
     level++;
     ignore=1;
   };
  if(strcmp(macro_command,"MACROEND/")==0)
   {
     level--;
     if(level==0)ignore=2;
   };
  };
/*********************************************************************/

 /* Now write the actual code itself */
                     /* only write out if not null statement and not ignoring */
  if(actual_statement[0]!='\0' && ignore==0)
   {                 /* write out statement */
      mjfputs (actual_statement); 
                     /* if any errors with include file write out error */
      if(err_flag==1)mjfputs(file_err);  
   }
                     /* clear ignore flag for one time only ignore */
  if(ignore==2)ignore=0;

/*********************************************************************/

};    /* MAIN LOOP END */

                    /* Setup env var include with # distinct
                       include files */
printf("includes=%-d\n",inccount);
printf("export includes\n");
};









/* PROCESS FILE FOR SAVING */
/*************************************************************************/
 void process_s()
/*************************************************************************/
{


while(inccount>=0)
{

 while(result!=0)
 {
  result=next_statement(inccount);

/*********************************************************************/
  /* If debugging and inside 4gl code block write debug stuff */

  if(debug_flag>0 && countlev[level]>0)
  {
   mjfputs("/*\n #debugbeg */\n");

   if(debug_flag==2)
   {
    mjfputs(";if ");
    mjfputs(condit_dbg);
    mjfputs(" then \n");
   };


   

   mjfputs(";message\n");
   mjfputs(print_statement);

   if(watchcnt>0 && popmode==1)
   {
    mjfputs("\'\n+X\'0D\'+\'");
    mjfputs(
"----------------------------WATCHED VARIABLES-----------------------------");
    mjfputs("\'+X\'0D\'\n");
    for (i=0; i<watchcnt; i++)  
    {
     mjfputs("+\' ");
     mjfputs(&watchlist[i][0]);


     mjfputs("=\'+ifnull(varchar(");
     mjfputs(&watchlist[i][0]);
     mjfputs("),\'*NULL*\')\n");
    };
   };

   if(watchcnt==0)mjfputs("\'\n");
   if(popmode==1)mjfputs("with style=popup (startrow=1);\n");
   if(popmode==0)mjfputs(&sleep_string);           /* V1.4c */

   if(debug_flag==2)
   {
    mjfputs("endif; \n");
   };
   mjfputs("/* #debugend \n*/");
  }

/*********************************************************************/
  /* Now write the actual code itself */
  if(actual_statement[0]!='\0')
         mjfputs (actual_statement);


/* FIGURE BLOCK LEVEL READY FOR NEXT STATEMENT */
/*********************************************************************/
  /* beginning of block? */
  if( result==2||result==4)
    {
    if( strstr(current_statement,"= SELECT")!=NULL|
        strstr(current_statement,"=SELECT" )!=NULL|
        strstr(current_statement,"= REPEATED SELECT")!=NULL|
        strstr(current_statement,"=REPEATED SELECT" )!=NULL|
        strstr(current_statement,"DISPLAY SUBMENU" )!=NULL|
        strstr(current_statement,"RUN SUBMENU" )!=NULL)
      {
       level++;
       countlev[level]=0;
      }
    else
      {
       countlev[level]++;
      };
    };

/*********************************************************************/
  /* end of block? */
  if( result==3||result==5)
    {
       countlev[level]--;
       if(countlev[level]<0)
       {
        level--;
        if(level<0)level=0;
       };
    };




/*********************************************************************/
  /* Assumption that all macro commands are on their own line or
     if they share a line they come after any commands */
/*********************************************************************/

  if(result==10)
  {
/*********************************************************************/
/* Allow user to include another file in the source file */
/*********************************************************************/
  strupr(macro_command,macro_command);
  strcpy(macro_param_upper,macro_param);
  strupr(macro_param_upper,macro_param_upper);
  if(strcmp(macro_command,"INCLUDE")==0)
   {
         incfile[inccount+1]=fopen(macro_param,"r");
	 if( incfile[inccount+1] == NULL) 
	   {
	     perror(macro_param);
             mjfputs(file_err);  
	   }
         else
           {
	     inccount++;
	
             readpnt[inccount]=(&readbuf[inccount][0]);
             *readpnt[inccount]='\0';
 	     mjfputs("/*\n #macrobeg */\n");	
           };    
   };


/*********************************************************************/
/* Allow user to control whether debug messages popup or are simple
   1 line messages */
/*********************************************************************/
  if(strcmp(macro_command,"POPUP")==0)
   {
     if(strcmp(macro_param_upper,"ON")==0)popmode=1;
     if(strcmp(macro_param_upper,"OFF")==0)popmode=0;
   };


/*********************************************************************/
/* Allow the user to control the length of the sleep command after 
   non popup debug message */
/*********************************************************************/
  if(strcmp(macro_command,"SLEEP")==0)                    /* V1.4c */
   {                                                      /* V1.4c */
     if(strcmp(macro_param,"0")==0)                       /* V1.4c */
		strcpy(sleep_string,";");                 /* V1.4c */
     if(strcmp(macro_param_upper,"OFF")==0)                /* V1.4c */
		strcpy(sleep_string,";");                 /* V1.4c */
                                                          /* V1.4c */
     if(strcmp(macro_param,"1")==0)                       /* V1.4c */
		strcpy(sleep_string,";sleep 1;\n");       /* V1.4c */
     if(strcmp(macro_param,"2")==0)                       /* V1.4c */
		strcpy(sleep_string,";sleep 2;\n");       /* V1.4c */
     if(strcmp(macro_param,"3")==0)                       /* V1.4c */
		strcpy(sleep_string,";sleep 3;\n");       /* V1.4c */
     if(strcmp(macro_param,"4")==0)                       /* V1.4c */
		strcpy(sleep_string,";sleep 4;\n");       /* V1.4c */
     if(strcmp(macro_param,"5")==0)                       /* V1.4c */
		strcpy(sleep_string,";sleep 5;\n");       /* V1.4c */
   };                                                     /* V1.4c */


/*********************************************************************/
/* Allow user to turn debugging on/off or conditional */
/*********************************************************************/
  if(strcmp(macro_command,"DEBUG")==0)
   {
     debug_flag=2;
     strcpy(condit_dbg,macro_param);

     if(strcmp(macro_param_upper,"ON")==0)debug_flag=1;
     if(strcmp(macro_param_upper,"OFF")==0)debug_flag=0;
   };


/*********************************************************************/
/* Allow user to watch specific 4gl variables */
/*********************************************************************/
  if(strcmp(macro_command,"WATCH")==0)
   {
    watchok=1;
    /* Check to see if already watching this variable */
    if(watchcnt>0)
    {
     for (i=0; i<watchcnt; i++)  
     {
      if(strcmp(macro_param,&watchlist[i][0])==0)watchok=0;
     };
    };

    if(watchok==1)
    {
     strcpy(&watchlist[watchcnt][0],macro_param);   
     watchcnt++;
    };
   };

/*********************************************************************/
/* Allow user to stop watching specifig 4gl variable */
/*********************************************************************/
  if(strcmp(macro_command,"NOWATCH")==0)
   {
    if(watchcnt>0)
    {
     for (i=0; i<watchcnt; i++)  
     {
      if(strcmp(macro_param,&watchlist[i][0])==0)
       {
        for (j=i; j<watchcnt; j++)
        {
         strcpy(&watchlist[j][0],&watchlist[j+1][0]);
        };
        watchcnt--;
       };
     };
    };
   };


/*********************************************************************/
  };
 };


/* If we are not on last file write macro end label */
if(inccount >0)mjfputs("\n/*#macroend \n*/");

/* End of file so drop down to previous file pointer */
current_char='\0';
inccount--;
result=1;
};
};






/**************************************************************************/
/* Get 1 statement or 1 macro comment from input file */
/**************************************************************************/
int next_statement( filenum)
int filenum;
{
char 	*current_ptr,*actual_ptr,*word_ptr,*print_ptr,*macro_ptr;
char    *test_ptr;

char 	last_char;

int     charcnt,print_lines;

                                  
last_char='\0';                     /* Initialize everything */
current_ptr=current_statement;
actual_ptr=actual_statement;
print_ptr=print_statement;
word_ptr=current_word;
macro_ptr=macro_command;

print_lines=0;

quote_flag=0;
comment_flag=0;                  
paren_level=0;                     /* V1.3c */
hash_flag=0;
charcnt=2;


*actual_ptr='\0';
*current_ptr='\0';
*word_ptr='\0';
*macro_ptr='\0';
macro_param[0]='\0';

*print_ptr++='\'';
*print_ptr='\0';
				    /* If last character was anything
				       but EOF clear it for next statement */	

if(current_char!=EOF)current_char='\0';

 
while (current_char!=EOF)           /* Only loop if not EOF */
 {
  last_char=current_char;	    /* Save last character */
  current_char=mjgetc(filenum);     /* Get next character */
  
  if(current_char!=EOF)             /* If not EOF process it */
  {

   /* save actual statement  including comments etc. */
   *actual_ptr++=current_char;
   *actual_ptr='\0';

   /* If debugging build up debug message here */
   if(debug_flag>0)
   { 


   /* But only 22 lines of debug message allowed! */	
   if((print_lines+watchcnt)<22 && 
      (popmode==1 || (popmode==0 && comment_flag==0))
     )
   {

    *print_ptr++=current_char;     /* Save character */
    *print_ptr='\0';
    charcnt++;                     /* Increment count of characters
				      on one line of source file for 
				      debug message statement */


    if(current_char=='\n')         /* If new line .. */
    {
    *--print_ptr='\0';             /* Ignore this character */

    /* V1.4c */
    print_ptr--;	           /* And point to last character */

     /* scan print statement so far to see if it is only spaces or tabs
       if so then ignore new line.  i.e don't write blank line as first
       line. 
     */

     test_ptr=print_statement;                   /* V1.4c */
     test_ptr++;                                 /* V1.4c */
     for (;                                      /* V1.4c */
          *test_ptr==' '||*test_ptr=='\t';       /* V1.4c */
          test_ptr++);                           /* V1.4c */
                                                 /* V1.4c */

     if(*test_ptr=='\0')            /* Was this blank line */ /* V1.4c */
      {
      charcnt=2;                   /* If so ignore whole line */
      print_ptr=print_statement;
      *print_ptr++='\'';
      *print_ptr='\0';
      print_lines=0;
      }
    else
     {
     print_ptr++;                  /* Otherwise newline */
                                   /* will be embedded in message  */
                                   /* statement */

     strcat(print_ptr,"\'+X\'0D\'+\'");
     print_ptr=print_ptr+9;
     charcnt=charcnt+9;
     print_lines++;                 /* Increment number of lines used */


                                                 /* V1.4c */
     /* If printing messages without popup limit message statement 
        length to 1 line (since only 1 line printed) by forcing loop
        to think that 22 print lines have been accumulated
     */
     if(popmode==0)                              /* V1.4c */
      {                                          /* V1.4c */
       print_lines=22;                           /* V1.4c */
      };                                         /* V1.4c */
     };
    };

    
    if(current_char=='\'')        /* If quotation */
    {                             /* Need to handle this specially */
    *--print_ptr='\0';            /* in message statement */
    strcat(print_ptr,"\'+X\'27\'+\'");
    print_ptr=print_ptr+9;
    charcnt=charcnt+9;
    };



    if(charcnt>65)                /* If we are over 65 characters then */
    {
    if(*--print_ptr=='\'')        /* If line has extra unbalanced quote */
     {
      *print_ptr='\0';            /* clear out extra quote */
     }
    else
     {
     print_ptr++;                 /* otherwise add in an extra quote */
     strcat(print_ptr,"\'+");
     print_ptr=print_ptr+2;
     };
    strcat(print_ptr,"\n\'");     /* Add on end of line plus new start quote */
    print_ptr=print_ptr+2;
    charcnt=2;
    };
   };
   };

   /* force tab or newline to space delimiter */
   if(current_char=='\n'||current_char=='\t')current_char=' ';
  };


  /* look for macro command inside comment */
  /* Note comment flag test put in outside 'if' for optimization */
  if(comment_flag==1)
  {
  if(current_char!=' ' && current_char!='*' &&
     current_char!='\t' && current_char!='\n')
  {
   switch (hash_flag)
   {
   case 3:
      if(current_char=='}')hash_flag=4;
      if(current_char!='}')
       {
         *macro_ptr++=current_char;
         *macro_ptr='\0';
       };
      break;
   case 2:
      if(current_char=='>')hash_flag=4;
      if(current_char!='>')
       { 
         *macro_ptr++=current_char;
         *macro_ptr='\0';
       };
      break;

   case 1:
      if(current_char=='<')
       {
	hash_flag=2;
	macro_ptr=macro_param;
        *macro_ptr='\0';
       };
      if(current_char=='{')
       {
        hash_flag=3;
        macro_ptr=macro_param;
        *macro_ptr='\0';
       };
      if(current_char!='<' && current_char!='{')
       {
         *macro_ptr++=current_char;
         *macro_ptr='\0';
       };
       break;

   case 0:
      macro_ptr=macro_command;
      *macro_ptr='\0';
      if(current_char=='#')hash_flag=1;
      break;
   };
  };
  };


  /* start of comment! */
  if(last_char=='/')
  {
   if(quote_flag==0 && current_char=='*')
     {
      *--current_ptr='\0';
      comment_flag=1;

     if(debug_flag>0 && popmode==0 )
     {
     test_ptr=print_statement;                   /* V1.4c */
     test_ptr++;                                 /* V1.4c */
     for (;                                      /* V1.4c */
          *test_ptr==' '||*test_ptr=='\t';       /* V1.4c */
          test_ptr++);                           /* V1.4c */ 
     if (*test_ptr++ =='/' && *test_ptr=='*')
      {
      charcnt=2;                   /* If so ignore whole line */
      print_ptr=print_statement;
      *print_ptr++='\'';
      *print_ptr='\0';
      print_lines=0;
      };
     };

      continue;
     };
  };
  
  /* end of comment? */
  if(current_char=='/')
  {
   if(last_char=='*' && quote_flag==0 ) /* Yes it really was */
     {
      comment_flag=0;
      if(hash_flag>0)return 10;  /* Possible macro command so check it */	
      continue;
     };
  };

  /* quotation mark */
  if(current_char=='\'')
  {
   if(comment_flag==0)
    {
    quote_flag=1-quote_flag;
    continue;
    };
  };


  /* pair of spaces suppress to one space*/
  if(current_char==' ' &&
     last_char==' ')continue;


  /* only look at data if not in quotes or comment */
  if(quote_flag==0 && comment_flag==0)
  {

    if(current_char!=EOF)
    {
     /* store testing data */
     *current_ptr++=current_char;
     *current_ptr='\0';
    };

   if(current_char<65 || 
      current_char>90)
   { 

    /* delimiter so must be new word */
    if(current_char==' '||
       current_char=='='||
       current_char==';'||
       current_char==','||
       current_char=='/'||
       current_char==')'||
       current_char=='('||	
       current_char==':'||
       current_char=='-'||
       current_char=='+'||
       current_char=='{'||
       current_char=='}'||
       current_char=='\\'||
       current_char=='>'||
       current_char=='<'||
       current_char=='&'||
       current_char=='*'||
       current_char==EOF)
    {
     strupr(current_word,current_word);
     /* check for key words that end statement */
     if(strcmp(current_word,"BEGIN")==0)return 4;
     if(strcmp(current_word,"END")==0)return 5;
     if(strcmp(current_word,"THEN")==0)return 6;
     if(strcmp(current_word,"DO")==0)return 7; /* V1.3c */
     if(strcmp(current_word,"ELSE")==0)return 8; /* V1.3c */

     /* Update current parenthesis nesting level V1.3c */
     if(current_char=='(')paren_level++;
     if(current_char==')'&&paren_level!=0)paren_level--;

     /* Only check for end of statement if we are not in parentheses V1.3c */
     if(paren_level==0)
      {
        /* is it a character that ends statement */
        if(current_char==';')return 1;
        if(current_char=='{')return 2;
        if(current_char=='}')return 3;
      };


     word_ptr=current_word;
     continue;
    };
   };
    *word_ptr++=current_char;
    *word_ptr='\0';
  };
 };

return 0;
}


/**************************************************************************/
/* Read 1 character from file whos id is stored in
   incfile[filenum]
*/
/**************************************************************************/

char mjgetc(filenum)
int filenum;
{
 char reslt;
 reslt=fgetc(incfile[filenum]);
 if(reslt==EOF)fclose(incfile[filenum]);
 return reslt;
}



/**************************************************************************/
/* Puts a string of characters in output buffer.  If the buffer
   overflows write out buffer to file. 
*/
/**************************************************************************/
int  mjfputs (outstring)
char *outstring;
{
int err=0;
fputs(outstring,outfile);
return err;
}

/**************************************************************************/
/* Flush out any characters left in the output buffer to file */
/**************************************************************************/
int mjflush ()
{
  int err=0;
  return err;
}

void strupr (outstring,instring)
char *outstring,*instring;
{
   while(*instring!='\0')
  {
    *outstring=*instring++;
    if(*outstring>=97 && *outstring<=122)*outstring=*outstring-32; 
    outstring++;
  } 
  *outstring='\0';
}
